from pathlib import Path
import numpy as np
import cv2


def imread(path: Path | str) -> np.ndarray:
    img = cv2.imread(str(path))
    assert img is not None, f'Cannot read from {path}'
    return img[:, :, ::-1].copy()


def imwrite(path: Path | str, image: np.ndarray) -> None:
    # Validate the path argument
    if not isinstance(path, (str, Path)):
        raise ValueError("Invalid 'path' argument. Expected a string or Path object.")
    path = Path(path)
    if not path.parent.exists():
        raise ValueError("Invalid 'path' argument. Parent directory does not exist.")
    if path.is_dir():
        raise ValueError("Invalid 'path' argument. Expected a file path, not a directory.")

    # Validate the image array
    if not isinstance(image, np.ndarray):
        raise ValueError("Invalid 'image' argument. Expected a NumPy array.")
    if len(image.shape) != 3 or image.shape[2] != 3:
        raise ValueError("Invalid 'image' argument. Expected a 3-dimensional array with shape (height, width, 3).")
    if not np.issubdtype(image.dtype, np.uint8):
        raise ValueError("Invalid 'image' argument. Expected a NumPy array with data type np.uint8.")

    # Save the image
    cv2.imwrite(str(path), image[:, :, ::-1].copy())



'''
Functions copied from python heapq, but decorated with numba.njit.
Also added arg- versions of these functions, to use in argtopk
'''

import numpy as np
import numba as nb


@nb.jit(nopython=True)
def _siftdown(heap, startpos, pos):
    newitem = heap[pos]
    # Follow the path to the root, moving parents down until finding a place
    # newitem fits.
    while pos > startpos:
        parentpos = (pos - 1) >> 1
        parent = heap[parentpos]
        if newitem < parent:
            heap[pos] = parent
            pos = parentpos
            continue
        break
    heap[pos] = newitem


@nb.jit(nopython=True)
def _siftup(heap, pos):
    endpos = len(heap)
    startpos = pos
    newitem = heap[pos]
    # Bubble up the smaller child until hitting a leaf.
    childpos = 2*pos + 1    # leftmost child position
    while childpos < endpos:
        # Set childpos to index of smaller child.
        rightpos = childpos + 1
        if rightpos < endpos and not heap[childpos] < heap[rightpos]:
            childpos = rightpos
        # Move the smaller child up.
        heap[pos] = heap[childpos]
        pos = childpos
        childpos = 2*pos + 1
    # The leaf at pos is empty now.  Put newitem there, and bubble it up
    # to its final resting place (by sifting its parents down).
    heap[pos] = newitem
    _siftdown(heap, startpos, pos)


@nb.jit(nopython=True)
def heappush(heap, item):
    """Push item onto heap, maintaining the heap invariant."""
    heap.append(item)
    _siftdown(heap, 0, len(heap)-1)


@nb.jit(nopython=True)
def heappop(heap):
    """Pop the smallest item off the heap, maintaining the heap invariant."""
    lastelt = heap.pop()    # raises appropriate IndexError if heap is empty
    if heap:
        returnitem = heap[0]
        heap[0] = lastelt
        _siftup(heap, 0)
        return returnitem
    return lastelt


@nb.jit(nopython=True)
def heapreplace(heap, item):
    """Pop and return the current smallest value, and add the new item.

    This is more efficient than heappop() followed by heappush(), and can be
    more appropriate when using a fixed-size heap.  Note that the value
    returned may be larger than item!  That constrains reasonable uses of
    this routine unless written as part of a conditional replacement:

        if item > heap[0]:
            item = heapreplace(heap, item)
    """
    returnitem = heap[0]    # raises appropriate IndexError if heap is empty
    heap[0] = item
    _siftup(heap, 0)
    return returnitem


@nb.jit(nopython=True)
def heappushpop(heap, item):
    """Fast version of a heappush followed by a heappop."""
    if heap[0] < item:
        item, heap[0] = heap[0], item
        _siftup(heap, 0)
    return item


@nb.jit(nopython=True)
def heapify(x):
    """Transform list into a heap, in-place, in O(len(x)) time."""
    n = len(x)
    # Transform bottom-up.  The largest index there's any point to looking at
    # is the largest with a child index in-range, so must have 2*i + 1 < n,
    # or i < (n-1)/2.  If n is even = 2*j, this is (2*j-1)/2 = j-1/2 so
    # j-1 is the largest, which is n//2 - 1.  If n is odd = 2*j+1, this is
    # (2*j+1-1)/2 = j so j-1 is the largest, and that's again n//2-1.
    for i in range(n // 2 - 1, -1, -1):
        _siftup(x, i)


@nb.jit(nopython=True)
def _argsiftdown(argheap, values, startpos, pos):
    newitem = argheap[pos]
    # Follow the path to the root, moving parents down until finding a place
    # newitem fits.
    while pos > startpos:
        parentpos = (pos - 1) >> 1
        parent = argheap[parentpos]
        if values[newitem] < values[parent]:
            argheap[pos] = parent
            pos = parentpos
            continue
        break
    argheap[pos] = newitem


@nb.jit(nopython=True)
def _argsiftup(argheap, values, pos):
    endpos = len(argheap)
    startpos = pos
    newitem = argheap[pos]
    # Bubble up the smaller child until hitting a leaf.
    childpos = 2*pos + 1    # leftmost child position
    while childpos < endpos:
        # Set childpos to index of smaller child.
        rightpos = childpos + 1
        if rightpos < endpos and not values[argheap[childpos]] < values[argheap[rightpos]]:
            childpos = rightpos
        # Move the smaller child up.
        argheap[pos] = argheap[childpos]
        pos = childpos
        childpos = 2*pos + 1
    # The leaf at pos is empty now.  Put newitem there, and bubble it up
    # to its final resting place (by sifting its parents down).
    argheap[pos] = newitem
    _argsiftdown(argheap, values, startpos, pos)


@nb.jit(nopython=True)
def argheappush(argheap, values, item):
    """Push item onto heap, maintaining the heap invariant."""
    argheap.append(item)
    _argsiftdown(argheap, values, 0, len(argheap) - 1)


@nb.jit(nopython=True)
def argheappop(argheap, values):
    """Pop the smallest item off the heap, maintaining the heap invariant."""
    lastelt = argheap.pop()    # raises appropriate IndexError if heap is empty
    if argheap:
        returnitem = argheap[0]
        argheap[0] = lastelt
        _argsiftup(argheap, values, 0)
        return returnitem
    return lastelt


@nb.jit(nopython=True)
def argheapreplace(argheap, values, item):
    """Pop and return the current smallest value, and add the new item.

    This is more efficient than heappop() followed by heappush(), and can be
    more appropriate when using a fixed-size heap.  Note that the value
    returned may be larger than item!  That constrains reasonable uses of
    this routine unless written as part of a conditional replacement:

        if item > heap[0]:
            item = heapreplace(heap, item)
    """
    returnitem = argheap[0]    # raises appropriate IndexError if heap is empty
    argheap[0] = item
    _argsiftup(argheap, values, 0)
    return returnitem


@nb.jit(nopython=True)
def argheappushpop(argheap, values, item):
    """Fast version of a heappush followed by a heappop."""
    if values[argheap[0]] < values[item]:
        item, argheap[0] = argheap[0], item
        _argsiftup(argheap, values, 0)
    return item


@nb.jit(nopython=True)
def argheapify(x):
    """Transform list into a heap, in-place, in O(len(x)) time."""
    n = len(x)
    # Transform bottom-up.  The largest index there's any point to looking at
    # is the largest with a child index in-range, so must have 2*i + 1 < n,
    # or i < (n-1)/2.  If n is even = 2*j, this is (2*j-1)/2 = j-1/2 so
    # j-1 is the largest, which is n//2 - 1.  If n is odd = 2*j+1, this is
    # (2*j+1-1)/2 = j so j-1 is the largest, and that's again n//2-1.
    indices = np.arange(len(x))
    for i in range(n // 2 - 1, -1, -1):
        _argsiftup(indices, x, i)
    return indices



import pytest
import numpy as np

from ruslanio.sorting import topk, argtopk


def topk_np_simple(seq, k):
    return np.sort(seq)[-k:][::-1]

def argtopk_np_simple(seq, k):
    return np.argsort(seq)[-k:][::-1]


Ns = [1, 10, 100, 1000]
ks = [1, 2, 3, 10, 20, 30]

test_data = [
    (np.random.random(N), k)
    for k in ks
    for N in Ns
]

@pytest.mark.parametrize("seq,k", test_data)
def test_topk(seq, k):
    ref = topk_np_simple(seq, k)
    res = topk(seq, k)
    assert (ref == res).all()


@pytest.mark.parametrize("seq,k", test_data)
def test_argtopk(seq, k):
    ref = argtopk_np_simple(seq, k)
    res = argtopk(seq, k)
    assert (seq[ref] == seq[res]).all()



import asyncio
from asyncio import Queue, Future


class batched:
    def __init__(self, func, batch_size=16, timeout=0.5):
        self.func = func
        self.batch_size = batch_size
        self.timeout = timeout
        self.input_q = Queue()
        self.output_q = Queue()
        asyncio.create_task(self._batching_worker())
        asyncio.create_task(self._waiter_worker())
        self.batch = []

    async def _batching_worker(self):
        while True:
            timed_out = False
            try:
                item = await asyncio.wait_for(self.input_q.get(), self.timeout)
                self.batch.append(item)
            except TimeoutError:
                timed_out = True

            if len(self.batch) >= self.batch_size or (len(self.batch) != 0 and timed_out):
                inputs = [pair[0] for pair in self.batch]
                futures = [pair[1] for pair in self.batch]
                asyncio.create_task(self._task(inputs, futures))
                self.batch = []

    async def _task(self, inputs, futures):
        results = await self.func(inputs)
        await self.output_q.put((results, futures))

    async def _waiter_worker(self):
        while True:
            futures: list[Future]
            results, futures = await self.output_q.get()
            for (result, future) in zip(results, futures):
                future.set_result(result)

    async def __call__(self, item):
        future = Future()
        await self.input_q.put((item, future))
        return await future


async def main():
    # Example usage
    async def my_batched_function(item_id_batch):
        await asyncio.sleep(1)
        print(f'called for {item_id_batch[0]}..{item_id_batch[-1]}')
        return [f'result_for_{item_id}' for item_id in item_id_batch]


    f = batched(my_batched_function, batch_size=10)

    all_item_ids = range(99)

    results = await asyncio.gather(*[f(item_id) for item_id in all_item_ids])
    print(results)


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(main())


import typing as tp

import numpy as np


def flatten_list(l: tp.Any, times: int = -1):
    # Check exit conditions
    if times == 0:
        return l
    if not isinstance(l, list):
        if times >= 0:
            raise ValueError(f'{type(l)} is not list')
        else:
            return l

    # Flatten recursively
    new_l = []
    for sub_l in l:
        for item in sub_l:
            new_l.append(flatten_list(item, times - 1))
    return new_l


class GenWithLen:
    def __init__(self, gen: tp.Iterable, length: int):
        self.gen = gen
        self.length = length

    def __iter__(self):
        for item in self.gen:
            yield item
    
    def __len__(self):
        return self.length
    
    def len(self):
        return self.length


def help_convolve_sliceable(source: tp.Iterable, kernel: int = 2, stride: int = 1):
    '''
    Generator function to perform a convolution-like operation on a sliceable input iterable.

    Parameters:
    - source (iterable): The input iterable to be processed (sliceable).
    - kernel (int, optional): The size of the sublists to yield (default is 2).
    - stride (int, optional): The step size when moving through the iterable (default is 1).
    
    Yields:
    - Slices: Slices of elements from the source iterable.

    This function takes a sliceable iterable, slices it into sublists of a specified size (kernel)
    with a specified step size (stride) while ensuring that stride is less than or equal to the kernel.
    '''
    assert kernel > 0
    assert stride > 0

    def _interator():
        # Start iterating over the source using slices
        for start in range(0, len(source) - kernel + stride, stride):
            end = start + kernel

            # Ensure that the end index does not exceed the length of the source
            end = min(end, len(source))

            # Yield the current slice
            yield source[start:end]
    
    return GenWithLen(_interator(), (len(source) - kernel + stride) // stride)


def help_convolve(source: tp.Iterable, kernel: int = 2, stride: int = 1):
    '''
    Generator function to perform a convolution-like operation on an input iterable.

    Parameters:
    - source (iterable): The input iterable to be processed.
    - kernel (int, optional): The size of the sublists to yield (default is 2).
    - stride (int, optional): The step size when moving through the iterable (default is 1).
    
    Yields:
    - List: Sublists of elements from the source iterable.

    This function takes an iterable, reads elements from it, and yields sublists of a specified
    size (kernel) with a specified step size (stride) while ensuring that stride is less than
    or equal to the kernel.
    '''

    # Check if stride is less than or equal to the kernel
    assert stride <= kernel
    assert kernel > 0
    assert stride > 0

    # Initialize an empty list to store elements from the source
    current = []

    # Create a generator from the source iterable
    gen = (item for item in source)

    # Infinite loop until a StopIteration exception is raised
    while True:
        try:
            # Get the next item from the generator
            item = next(gen)
        except StopIteration:
            # Break the loop when there are no more items in the generator
            break
        
        # Append the item to the current list
        current.append(item)

        # Check if the current list has reached the specified kernel size
        if len(current) == kernel:
            # Yield the current sublist as a result
            yield current

            # Update `current` to contain the last `kernel - stride` elements
            current = current[stride:]

    # If there are remaining elements in current (fewer than kernel elements), yield them
    # current = current[stride:]
    if len(current) > 0:
        yield current



import numpy as np
import numba as nb

from .fast_heapq import heapify, heappushpop
from .fast_heapq import argheapify, argheappushpop


@nb.jit(nopython=True)
def topk(seq: np.ndarray, k=10):
    '''
    Returns top-k values of the input array.
    Complexity is O(N * log(k)), where N = len(seq).
    Works only with numpy arrays.

    First time calling this function will be slow because it will compile.

    Equivalent to `np.sort(seq)[-k:][::-1]` or `heapq.nlargest(k, seq)`.
    First option is slow because its complexity is O(N * log(N)).
    Second option is slow because it is pure python.
    This function is fast because it is compiled using numba.jit.
    '''
    topheap = seq[:k].copy()
    heapify(topheap)

    for i in range(k, len(seq)):
        heappushpop(topheap, seq[i])

    return np.sort(topheap)[::-1]


@nb.jit(nopython=True)
def argtopk(seq: np.ndarray, k=10):
    '''
    Works like `ruslanio.sorting.topk`, but returns indeces instead of values.

    Equivalent to `np.argsort(seq)[-k:][::-1]`
    '''
    topheap = argheapify(seq[:k])

    for i in range(k, len(seq)):
        argheappushpop(topheap, seq, i)

    return topheap[np.argsort(seq[topheap])[::-1]]



import typing as tp
from abc import ABC, abstractmethod
import numpy as np

from .sorting import argtopk
from .arrays import flatten_list


class Sequence(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def updated_copy(self, item: tp.Any):
        pass

    @abstractmethod
    def get_candidates(self) -> list[tuple[tp.Any, float]]:
        pass

    @abstractmethod
    def ended(self) -> bool:
        pass


class NoItemClass:
    pass


NoItem = NoItemClass()


class BeamSearch:
    def __init__(self, beam_size: int):
        self.beam_size = beam_size

    def search(
        self,
        initial_sequences: list[Sequence],
        max_iters: int,
        initial_scores: list[float] | None = None,
        tqdm=None,
    ) -> Sequence:
        # Initialize beam
        sequences = initial_sequences
        if initial_scores is not None:
            scores = initial_scores
        else:
            scores = [0 for _ in sequences]

        pbar = range(max_iters)
        if tqdm is not None:
            pbar = tqdm(range(max_iters))
        for _ in pbar:
            # Generate new points
            candidates = [
                [(seq_i, item, delta_score) for (item, delta_score) in seq.get_candidates()]
                for (seq_i, seq) in enumerate(sequences) if not seq.ended()
            ]  # [n_candidates * current_beam_size]

            # Add stopped sequences
            candidates.extend([
                [(seq_i, NoItem, 0)]
                for (seq_i, seq) in enumerate(sequences) if seq.ended()
            ])

            # Find top scoring candidates
            candidates = flatten_list(candidates, 1)
            candidate_scores = np.array([
                scores[seq_i] + delta_score
                for (seq_i, _, delta_score) in candidates
            ])
            top_i = argtopk(candidate_scores, self.beam_size)

            # Update beam
            new_sequences = []
            for i in top_i:
                seq_i, item, _ = candidates[i]
                if item is not NoItem:
                    new_seq = sequences[seq_i].updated_copy(item)
                else:
                    new_seq = sequences[seq_i]
                new_sequences.append(new_seq)
            sequences = new_sequences

            # Update scores
            scores = candidate_scores[top_i]

            pbar.set_postfix(best_score=scores[0])
            
            if all([seq.ended() for seq in sequences]):
                break

        return sequences, scores
